"""
Module containing tests regarding the persistent_data memory and the ensemble output
"""

import time

import json
import textwrap
import binary_c_python_api




from mergedict import ConfigDict

# import unicode

class Decoder(json.JSONDecoder):
    """
    Custom decoder to transform the numbers that are strings to actual floats
    """

    def decode(self, s):
        result = super().decode(s)  # result = super(Decoder, self).decode(s) for Python 2.x
        return self._decode(result)

    def _decode(self, o):
        """
        Depending on the type of object is will determine whether to loop over the elements,
        or try to change the type of the object from string to float

        The try except might be a somewhat rough solution but it catches all cases.
        """


        # Check if we can turn it into a float
        # if isinstance(o, str) or isinstance(o, unicode):
        if isinstance(o, str):
            try:
                return float(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return {k: self._decode(v) for k, v in o.items()}
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o

# Evolution functions
def test_return_persistent_data_memaddr():
    output = binary_c_python_api.return_persistent_data_memaddr("")

    print("function: test_run_system")
    print("Binary_c output:")
    print(textwrap.indent(str(output), "\t"))

def test_passing_persistent_data_to_run_system():
    # Function to test the passing of the persistent data memoery adress, and having ensemble_defer = True
    # We should see that the results of multiple systems have been added to the one output json

    # Some values
    m1 = 15.0  # Msun
    m2 = 14.0  # Msun
    separation = 0  # 0 = ignored, use period
    orbital_period = 453000000000.0  # days
    eccentricity = 0.0
    metallicity = 0.02
    max_evolution_time = 15000

    # Make the argstrings
    argstring_template = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g}\
    eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g} ensemble 1 ensemble_defer {7}\
    ensemble_filters_off 1 ensemble_filter_SUPERNOVAE 1 probability 0.1"

    argstring_1 = argstring_template.format(m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "0")
    argstring_1_deferred = argstring_template.format(m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "1")
    argstring_2 = argstring_template.format(m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "0")

    # 
    persistent_data_memaddr = binary_c_python_api.return_persistent_data_memaddr()

    output_1 = binary_c_python_api.run_system(argstring=argstring_1)
    ensemble_jsons_1 = [line for line in output_1.splitlines() if line.startswith("ENSEMBLE_JSON")]
    json_1 = json.loads(ensemble_jsons_1[0][len("ENSEMBLE_JSON "):])

    # Doing 2 systems in a row.
    output_1_deferred = binary_c_python_api.run_system(argstring=argstring_1_deferred, persistent_data_memaddr=persistent_data_memaddr)
    output_2 = binary_c_python_api.run_system(argstring=argstring_2, persistent_data_memaddr=persistent_data_memaddr)
    ensemble_jsons_2 = [line for line in output_2.splitlines() if line.startswith("ENSEMBLE_JSON")]
    json_2 = json.loads(ensemble_jsons_2[0][len("ENSEMBLE_JSON "):])

    # Doing system one again.
    output_1_again = binary_c_python_api.run_system(argstring=argstring_1)
    ensemble_jsons_1 = [line for line in output_1_again.splitlines() if line.startswith("ENSEMBLE_JSON")]
    json_1_again = json.loads(ensemble_jsons_1[0][len("ENSEMBLE_JSON "):])

    print("Twice the same system are they equal?:")
    print(json_1 == json_1_again)

    print("first system vs deferred double system?:")
    print(json_1 == json_2)

    # printf("combined double system vs deferred double system?:")

def ensemble_output():

    m1 = 15.0  # Msun
    m2 = 14.0  # Msun
    separation = 0  # 0 = ignored, use period
    orbital_period = 453000000000.0  # days
    eccentricity = 0.0
    metallicity = 0.02
    max_evolution_time = 15000

    argstring_1 = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g} eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g} ensemble 1 ensemble_defer 0 ensemble_filters_off 1 ensemble_filter_SUPERNOVAE 1".format(
        m1,
        m2,
        separation,
        orbital_period,
        eccentricity,
        metallicity,
        max_evolution_time,
    )
    output_1 = binary_c_python_api.run_system(argstring=argstring_1)
    ensemble_jsons_1 = [line for line in output_1.splitlines() if line.startswith("ENSEMBLE_JSON")]

    start = time.time()
    json_1 = json.loads(ensemble_jsons_1[0][len("ENSEMBLE_JSON "):], cls=Decoder)
    stop = time.time()

    print(json.dumps(json_1, indent=4))
    print("took {}s to decode".format(stop-start))

def adding_ensemble_output():
    m1 = 2  # Msun
    m2 = 0.1  # Msun
    separation = 0  # 0 = ignored, use period
    orbital_period = 453000000.0  # days
    eccentricity = 0.0
    metallicity = 0.02
    max_evolution_time = 15000

    # Set up argstrings
    argstring_template = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g}\
    eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g} ensemble 1 ensemble_defer {7}\
    ensemble_filters_off 1 ensemble_filter_STELLAR_TYPE_COUNTS 1 probability 0.1"

    #############################################################################################
    # The 2 runs below use the ensemble but do not defer the output to anything else, so that the
    # results are returned directly after the run

    # Direct output commands
    argstring_1 = argstring_template.format(
        m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "0")
    argstring_2 = argstring_template.format(
        m1+1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "0")

    # Get outputs
    output_1 = binary_c_python_api.run_system(argstring=argstring_1)
    output_2 = binary_c_python_api.run_system(argstring=argstring_2)

    ensemble_jsons_1 = [line for line in output_1.splitlines() if line.startswith("ENSEMBLE_JSON")]
    ensemble_jsons_2 = [line for line in output_2.splitlines() if line.startswith("ENSEMBLE_JSON")]

    json_1 = json.loads(ensemble_jsons_1[0][len("ENSEMBLE_JSON "):], cls=Decoder)
    json_2 = json.loads(ensemble_jsons_2[0][len("ENSEMBLE_JSON "):], cls=Decoder)

    with open("json_1.json", 'w') as f:
        f.write(json.dumps(json_1, indent=4))
    with open("json_2.json", 'w') as f:
        f.write(json.dumps(json_2, indent=4))

    print("Single runs done\n")

    #############################################################################################
    # The 2 runs below use the ensemble and both defer the output so that after they are finished
    # nothing is printed. After that we explicitly free the memory of the persistent_data and 
    # have the output returned in that way

    # Deferred commands
    argstring_1_deferred = argstring_template.format(
        m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "1")
    argstring_2_deferred = argstring_template.format(
        m1+1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "1")

    # Get a memory location
    persistent_data_memaddr = binary_c_python_api.return_persistent_data_memaddr()

    # Run the systems and defer the output each time
    output_1_deferred = binary_c_python_api.run_system(argstring=argstring_1_deferred, persistent_data_memaddr=persistent_data_memaddr)
    output_2_deferred = binary_c_python_api.run_system(argstring=argstring_2_deferred, persistent_data_memaddr=persistent_data_memaddr)

    # Have the persistent_memory adress be released and have the json outputted
    output_total_deferred = binary_c_python_api.free_persistent_data_memaddr_and_return_json_output(persistent_data_memaddr)

    ensemble_jsons_deferred = [line for line in output_total_deferred.splitlines() if line.startswith("ENSEMBLE_JSON")]

    json_deferred = json.loads(ensemble_jsons_deferred[0][len("ENSEMBLE_JSON "):], cls=Decoder)

    with open("json_deferred.json", 'w') as f:
        f.write(json.dumps(json_deferred, indent=4))

    print("Double deferred done\n")

    #############################################################################################
    # The 2 runs below use the ensemble and the first one defers the output to the memory,
    # Then the second one uses that memory to combine its results with, but doesn't defer the 
    # data after that, so it will print it after the second run is done

    persistent_data_memaddr_2 = binary_c_python_api.return_persistent_data_memaddr()

    # Run the systems and defer the output once and the second time not, so that the second run automatically prints out the results
    output_1_deferred = binary_c_python_api.run_system(argstring=argstring_1_deferred, persistent_data_memaddr=persistent_data_memaddr_2)
    output_2_deferred_and_output = binary_c_python_api.run_system(argstring=argstring_2, persistent_data_memaddr=persistent_data_memaddr_2)

    ensemble_jsons_deferred_and_output = [line for line in output_2_deferred_and_output.splitlines() if line.startswith("ENSEMBLE_JSON")]

    json_deferred_and_output = json.loads(ensemble_jsons_deferred_and_output[0][len("ENSEMBLE_JSON "):], cls=Decoder)

    with open("json_deferred_and_output.json", 'w') as f:
        f.write(json.dumps(json_deferred_and_output, indent=4))

    print("Single deferred done\n")

def test_free_and_json_output():
    m1 = 2  # Msun
    m2 = 0.1  # Msun
    separation = 0  # 0 = ignored, use period
    orbital_period = 453000000000.0  # days
    eccentricity = 0.0
    metallicity = 0.02
    max_evolution_time = 15000

    # Get a memory adress:
    persistent_data_memaddr = binary_c_python_api.return_persistent_data_memaddr("")

    # Set up argstrings
    argstring_template = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g}\
    eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g} ensemble 1 ensemble_defer {7}\
    ensemble_filters_off 1 ensemble_filter_STELLAR_TYPE_COUNTS 1 probability 0.1"
    argstring_1 = argstring_template.format(
        m1, m2, separation, orbital_period, eccentricity, metallicity, max_evolution_time, "1")

    print("evolving")
    output_1_deferred = binary_c_python_api.run_system(argstring=argstring_1, persistent_data_memaddr=persistent_data_memaddr)
    print("Evolved")
    print("Output:")
    print(textwrap.indent(str(output_1_deferred), "\t"))

    print("freeing")
    json_output_by_freeing = binary_c_python_api.free_persistent_data_memaddr_and_return_json_output(persistent_data_memaddr)
    print("Freed")
    print("Output:")
    print(textwrap.indent(str(json_output_by_freeing), "\t"))


####
if __name__ == "__main__":
    # test_return_persistent_data_memaddr()
    # test_passing_persistent_data_to_run_system()
    # ensemble_output()
    adding_ensemble_output()




    # test_free_and_json_output()